MAX_TGW_CNT_PER_PBU = 10
XTP_VERSION_LEN = 16
XTP_TRADING_DAY_LEN = 9
XTP_TICKER_LEN = 16
XTP_TICKER_NAME_LEN = 64
XTP_LOCAL_ORDER_LEN = 11
XTP_ORDER_EXCH_LEN = 17
XTP_EXEC_ID_LEN = 18
XTP_BRANCH_PBU_LEN = 7
XTP_ACCOUNT_NAME_LEN = 16
XTP_CREDIT_DEBT_ID_LEN = 33
XTP_INET_ADDRESS_STR_LEN = 64
XTP_MAC_ADDRESS_LEN = 16
XTP_HARDDISK_SN_LEN = 24
XTP_MACOS_SNO_LEN = 21
XTP_STRATEGE_LEG_NUM = 4
XTP_STRATEGY_ID_LEN = 10
XTP_STRATEGY_NAME_LEN = 32
XTP_SECONDARY_ORDER_ID_LEN = 18
XTP_CNTRT_COMB_STRA_LIST_LEN = 2048
XTP_COMBINED_EXECUTION_LEG_NUM = 2
XTP_LOG_LEVEL_FATAL = 1
XTP_LOG_LEVEL_ERROR = 2
XTP_LOG_LEVEL_WARNING = 3
XTP_LOG_LEVEL_INFO = 4
XTP_LOG_LEVEL_DEBUG = 5
XTP_LOG_LEVEL_TRACE = 6
XTP_PROTOCOL_TCP = 1
XTP_PROTOCOL_UDP = 2
XTP_EXCHANGE_SH = 1
XTP_EXCHANGE_SZ = 2
XTP_EXCHANGE_UNKNOWN = 3
XTP_MKT_INIT = 0
XTP_MKT_SZ_A = 1
XTP_MKT_SH_A = 2
XTP_MKT_UNKNOWN = 3
XTP_PRICE_LIMIT = 1
XTP_PRICE_BEST_OR_CANCEL = 2
XTP_PRICE_BEST5_OR_LIMIT = 3
XTP_PRICE_BEST5_OR_CANCEL = 4
XTP_PRICE_ALL_OR_CANCEL = 5
XTP_PRICE_FORWARD_BEST = 6
XTP_PRICE_REVERSE_BEST_LIMIT = 7
XTP_PRICE_LIMIT_OR_CANCEL = 8
XTP_PRICE_TYPE_UNKNOWN = 9
XTP_SIDE_BUY = 1
XTP_SIDE_SELL = 2
XTP_SIDE_PURCHASE = 7
XTP_SIDE_REDEMPTION = 8
XTP_SIDE_SPLIT = 9
XTP_SIDE_MERGE = 10
XTP_SIDE_COVER = 11
XTP_SIDE_FREEZE = 12
XTP_SIDE_MARGIN_TRADE = 21
XTP_SIDE_SHORT_SELL = 22
XTP_SIDE_REPAY_MARGIN = 23
XTP_SIDE_REPAY_STOCK = 24
XTP_SIDE_STOCK_REPAY_STOCK = 26
XTP_SIDE_SURSTK_TRANS = 27
XTP_SIDE_GRTSTK_TRANSIN = 28
XTP_SIDE_GRTSTK_TRANSOUT = 29
XTP_SIDE_OPT_COMBINE = 31
XTP_SIDE_OPT_SPLIT = 32
XTP_SIDE_OPT_SPLIT_FORCE = 33
XTP_SIDE_OPT_SPLIT_FORCE_EXCH = 34
XTP_SIDE_UNKNOWN = 50
XTP_POSITION_EFFECT_INIT = 0
XTP_POSITION_EFFECT_OPEN = 1
XTP_POSITION_EFFECT_CLOSE = 2
XTP_POSITION_EFFECT_FORCECLOSE = 3
XTP_POSITION_EFFECT_CLOSETODAY = 4
XTP_POSITION_EFFECT_CLOSEYESTERDAY = 5
XTP_POSITION_EFFECT_FORCEOFF = 6
XTP_POSITION_EFFECT_LOCALFORCECLOSE = 7
XTP_POSITION_EFFECT_CREDIT_FORCE_COVER = 8
XTP_POSITION_EFFECT_CREDIT_FORCE_CLEAR = 9
XTP_POSITION_EFFECT_CREDIT_FORCE_DEBT = 10
XTP_POSITION_EFFECT_CREDIT_FORCE_UNCOND = 11
XTP_POSITION_EFFECT_UNKNOWN = 12
XTP_ORDER_ACTION_STATUS_SUBMITTED = 1
XTP_ORDER_ACTION_STATUS_ACCEPTED = 2
XTP_ORDER_ACTION_STATUS_REJECTED = 3
XTP_ORDER_STATUS_INIT = 0
XTP_ORDER_STATUS_ALLTRADED = 1
XTP_ORDER_STATUS_PARTTRADEDQUEUEING = 2
XTP_ORDER_STATUS_PARTTRADEDNOTQUEUEING = 3
XTP_ORDER_STATUS_NOTRADEQUEUEING = 4
XTP_ORDER_STATUS_CANCELED = 5
XTP_ORDER_STATUS_REJECTED = 6
XTP_ORDER_STATUS_UNKNOWN = 7
XTP_ORDER_SUBMIT_STATUS_INSERT_SUBMITTED = 1
XTP_ORDER_SUBMIT_STATUS_INSERT_ACCEPTED = 2
XTP_ORDER_SUBMIT_STATUS_INSERT_REJECTED = 3
XTP_ORDER_SUBMIT_STATUS_CANCEL_SUBMITTED = 4
XTP_ORDER_SUBMIT_STATUS_CANCEL_REJECTED = 5
XTP_ORDER_SUBMIT_STATUS_CANCEL_ACCEPTED = 6
XTP_TERT_RESTART = 0
XTP_TERT_RESUME = 1
XTP_TERT_QUICK = 2
XTP_TICKER_TYPE_STOCK = 0
XTP_TICKER_TYPE_INDEX = 1
XTP_TICKER_TYPE_FUND = 2
XTP_TICKER_TYPE_BOND = 3
XTP_TICKER_TYPE_OPTION = 4
XTP_TICKER_TYPE_TECH_STOCK = 5
XTP_TICKER_TYPE_UNKNOWN = 6
XTP_BUSINESS_TYPE_CASH = 0
XTP_BUSINESS_TYPE_IPOS = 1
XTP_BUSINESS_TYPE_REPO = 2
XTP_BUSINESS_TYPE_ETF = 3
XTP_BUSINESS_TYPE_MARGIN = 4
XTP_BUSINESS_TYPE_DESIGNATION = 5
XTP_BUSINESS_TYPE_ALLOTMENT = 6
XTP_BUSINESS_TYPE_STRUCTURED_FUND_PURCHASE_REDEMPTION = 7
XTP_BUSINESS_TYPE_STRUCTURED_FUND_SPLIT_MERGE = 8
XTP_BUSINESS_TYPE_MONEY_FUND = 9
XTP_BUSINESS_TYPE_OPTION = 10
XTP_BUSINESS_TYPE_EXECUTE = 11
XTP_BUSINESS_TYPE_FREEZE = 12
XTP_BUSINESS_TYPE_OPTION_COMBINE = 13
XTP_BUSINESS_TYPE_EXECUTE_COMBINE = 14
XTP_BUSINESS_TYPE_UNKNOWN = 15
XTP_ACCOUNT_NORMAL = 0
XTP_ACCOUNT_CREDIT = 1
XTP_ACCOUNT_DERIVE = 2
XTP_ACCOUNT_UNKNOWN = 3
XTP_FUND_TRANSFER_OUT = 0
XTP_FUND_TRANSFER_IN = 1
XTP_FUND_INTER_TRANSFER_OUT = 2
XTP_FUND_INTER_TRANSFER_IN = 3
XTP_FUND_TRANSFER_UNKNOWN = 4
XTP_FUND_OPER_PROCESSING = 0
XTP_FUND_OPER_SUCCESS = 1
XTP_FUND_OPER_FAILED = 2
XTP_FUND_OPER_SUBMITTED = 3
XTP_FUND_OPER_UNKNOWN = 4
XTP_DEBT_EXTEND_OPER_PROCESSING = 0
XTP_DEBT_EXTEND_OPER_SUBMITTED = 1
XTP_DEBT_EXTEND_OPER_SUCCESS = 2
XTP_DEBT_EXTEND_OPER_FAILED = 3
XTP_DEBT_EXTEND_OPER_UNKNOWN = 4
XTP_SPLIT_MERGE_STATUS_ALLOW = 0
XTP_SPLIT_MERGE_STATUS_ONLY_SPLIT = 1
XTP_SPLIT_MERGE_STATUS_ONLY_MERGE = 2
XTP_SPLIT_MERGE_STATUS_FORBIDDEN = 3
XTP_TBT_ENTRUST = 1
XTP_TBT_TRADE = 2
XTP_OPT_CALL = 1
XTP_OPT_PUT = 2
XTP_OPT_EXERCISE_TYPE_EUR = 1
XTP_OPT_EXERCISE_TYPE_AME = 2
XTP_POSITION_DIRECTION_NET = 0
XTP_POSITION_DIRECTION_LONG = 1
XTP_POSITION_DIRECTION_SHORT = 2
XTP_POSITION_DIRECTION_COVERED = 3
XTP_POSITION_UNCOVERED = 0
XTP_POSITION_COVERED = 1
XTP_CRD_CR_INIT = 0
XTP_CRD_CR_SUCCESS = 1
XTP_CRD_CR_FAILED = 2
XTP_OPT_POSITION_TYPE_CONTRACT = 0
XTP_OPT_POSITION_TYPE_COMBINED = 1
XTP_ORDER_DETAIL_TYPE_NEW_ORDER = 0
XTP_ORDER_DETAIL_TYPE_CANCEL_ORDER = 1
XTP_ORDER_DETAIL_TYPE_OPT_COMB_NEW_ORDER = 2
XTP_ORDER_DETAIL_TYPE_OPT_COMB_CANCEL_ORDER = 3
XTP_TRDT_COMMON = '0'
XTP_TRDT_CASH = '1'
XTP_TRDT_PRIMARY = '2'
XTP_TRDT_CROSS_MKT_CASH = '3'
XTP_ORDT_Normal = '0'
XTP_ORDT_DeriveFromQuote = '1'
XTP_ORDT_DeriveFromCombination = '2'
XTP_ORDT_Combination = '3'
XTP_ORDT_ConditionalOrder = '4'
XTP_ORDT_Swap = '5'
XTP_TERMINAL_PC = 1
XTP_TERMINAL_ANDROID = 2
XTP_TERMINAL_IOS = 3
XTP_TERMINAL_WP = 4
XTP_TERMINAL_STATION = 5
XTP_TERMINAL_TEL = 6
XTP_TERMINAL_PC_LINUX = 7
XTP_EXP_DATE_SAME = 0
XTP_EXP_DATE_DIFF = 1
XTP_EXP_DATE_NON = 2
XTP_UNDERLYING_SAME = 0
XTP_UNDERLYING_DIFF = 1
XTP_UNDERLYING_NON = 2
XTP_AUTO_SPLIT_EXPDAY = 0
XTP_AUTO_SPLIT_PREDAY = 1
XTP_AUTO_SPLIT_PRE2DAY = 2
XTP_AUTO_SPLIT_NON = 3
XTP_QUALIFICATION_PUBLIC = 0
XTP_QUALIFICATION_COMMON = 1
XTP_QUALIFICATION_ORGANIZATION = 2
XTP_QUALIFICATION_UNKNOWN = 3
XTP_SECURITY_MAIN_BOARD = 0
XTP_SECURITY_SECOND_BOARD = 1
XTP_SECURITY_STARTUP_BOARD = 2
XTP_SECURITY_INDEX = 3
XTP_SECURITY_TECH_BOARD = 4
XTP_SECURITY_STATE_BOND = 5
XTP_SECURITY_ENTERPRICE_BOND = 6
XTP_SECURITY_COMPANEY_BOND = 7
XTP_SECURITY_CONVERTABLE_BOND = 8
XTP_SECURITY_NATIONAL_BOND_REVERSE_REPO = 12
XTP_SECURITY_ETF_SINGLE_MARKET_STOCK = 14
XTP_SECURITY_ETF_INTER_MARKET_STOCK = 15
XTP_SECURITY_ETF_SINGLE_MARKET_BOND = 17
XTP_SECURITY_ETF_GOLD = 19
XTP_SECURITY_STRUCTURED_FUND_CHILD = 24
XTP_SECURITY_SZSE_RECREATION_FUND = 26
XTP_SECURITY_STOCK_OPTION = 29
XTP_SECURITY_ETF_OPTION = 30
XTP_SECURITY_ALLOTMENT = 100
XTP_SECURITY_MONETARY_FUND_SHCR = 110
XTP_SECURITY_MONETARY_FUND_SHTR = 111
XTP_SECURITY_MONETARY_FUND_SZ = 112
XTP_SECURITY_OTHERS = 255



import pytz
from typing import Any, Dict, List
from datetime import datetime
from copy import copy

from vnpy.event import EventEngine
from vnpy.trader.event import EVENT_TIMER
from vnpy.trader.constant import (
    Exchange,
    Product,
    Direction,
    OrderType,
    Status,
    Offset,
    OptionType
)
from vnpy.trader.gateway import BaseGateway
from vnpy.trader.object import (
    CancelRequest,
    OrderRequest,
    SubscribeRequest,
    TickData,
    ContractData,
    OrderData,
    TradeData,
    PositionData,
    AccountData
)
from vnpy.trader.utility import get_folder_path, round_to

from vnpy_xtp.api import MdApi, TdApi


# 交易所映射
MARKET_XTP2VT: Dict[int, Exchange] = {
    1: Exchange.SZSE,
    2: Exchange.SSE
}
MARKET_VT2XTP: Dict[Exchange, int] = {v: k for k, v in MARKET_XTP2VT.items()}

EXCHANGE_XTP2VT: Dict[int, Exchange] = {
    1: Exchange.SSE,
    2: Exchange.SZSE,
}
EXCHANGE_VT2XTP: Dict[Exchange, int] = {v: k for k, v in EXCHANGE_XTP2VT.items()}

# 多空和开平方向映射
DIRECTION_STOCK_XTP2VT: Dict[int, Any] = {
    1: (Direction.LONG, Offset.NONE),
    2: (Direction.SHORT, Offset.NONE),
    21: (Direction.LONG, Offset.OPEN),
    22: (Direction.SHORT, Offset.OPEN),
    24: (Direction.LONG, Offset.CLOSE),
    23: (Direction.SHORT, Offset.CLOSE)
}
DIRECTION_STOCK_VT2XTP: Dict[Any, int] = {v: k for k, v in DIRECTION_STOCK_XTP2VT.items()}

DIRECTION_OPTION_XTP2VT: Dict[int, Direction] = {
    1: Direction.LONG,
    2: Direction.SHORT
}
DIRECTION_OPTION_VT2XTP: Dict[Direction, int] = {v: k for k, v in DIRECTION_OPTION_XTP2VT.items()}

# 持仓方向映射
POSITION_DIRECTION_XTP2VT = {
    0: Direction.NET,
    1: Direction.LONG,
    2: Direction.SHORT,
    3: Direction.SHORT
}

# 期权委托类型映射
OPTION_ORDERTYPE_XTP2VT: Dict[int, OrderType] = {
    1: OrderType.LIMIT,
    2: OrderType.MARKET,
    8: OrderType.FOK
}
OPTION_ORDERTYPE_VT2XTP: Dict[OrderType, int] = {v: k for k, v in OPTION_ORDERTYPE_XTP2VT.items()}

# 股票委托类型映射
EQUITY_ORDERTYPE_XTP2VT: Dict[int, OrderType] = {
    1: OrderType.LIMIT,
    4: OrderType.MARKET
}
EQUITY_ORDERTYPE_VT2XTP: Dict[OrderType, int] = {v: k for k, v in EQUITY_ORDERTYPE_XTP2VT.items()}

# 科创板委托类型映射
STAR_ORDERTYPE_XTP2VT: Dict[int, OrderType] = {
    1: OrderType.LIMIT,
    7: OrderType.MARKET
}
STAR_ORDERTYPE_VT2XTP: Dict[OrderType, int] = {v: k for k, v in STAR_ORDERTYPE_XTP2VT.items()}

# 通讯协议映射
PROTOCOL_VT2XTP: Dict[str, int] = {
    "TCP": 1,
    "UDP": 2
}

# 委托状态映射
STATUS_XTP2VT: Dict[int, Status] = {
    0: Status.SUBMITTING,
    1: Status.ALLTRADED,
    2: Status.PARTTRADED,
    3: Status.CANCELLED,
    4: Status.NOTTRADED,
    5: Status.CANCELLED,
    6: Status.REJECTED,
    7: Status.SUBMITTING
}

# 产品类型映射
PRODUCT_XTP2VT: Dict[int, Product] = {
    0: Product.EQUITY,
    1: Product.INDEX,
    2: Product.FUND,
    3: Product.BOND,
    4: Product.OPTION,
    5: Product.EQUITY,
    6: Product.FUND
}

# 开平方向映射
OFFSET_VT2XTP: Dict[Offset, int] = {
    Offset.NONE: 0,
    Offset.OPEN: 1,
    Offset.CLOSE: 2,
    Offset.CLOSETODAY: 4,
    Offset.CLOSEYESTERDAY: 5
}
OFFSET_XTP2VT: Dict[int, Offset] = {v: k for k, v in OFFSET_VT2XTP.items()}

# 业务类型映射
BUSINESS_VT2XTP: Dict[Any, int] = {
    "CASH": 0,
    Offset.NONE: 0,
    "MARGIN": 4,
    Offset.OPEN: 4,
    Offset.CLOSE: 4,
    "OPTION": 10,
}

# 期权类型映射
OPTIONTYPE_XTP2VT = {
    1: OptionType.CALL,
    2: OptionType.PUT
}

# 日志级别映射
LOGLEVEL_VT2XTP = {
    "FATAL": 0,
    "ERROR": 1,
    "WARNING": 2,
    "INFO": 3,
    "DEBUG": 4,
    "TRACE": 5,
}

# 时区常量
CHINA_TZ = pytz.timezone("Asia/Shanghai")

# 合约数据全局缓存字典
symbol_contract_map: Dict[str, ContractData] = {}


class XtpGateway(BaseGateway):
    """
    vn.py用于对接中泰XTP柜台的交易接口。
    """

    default_setting: Dict[str, Any] = {
        "账号": "",
        "密码": "",
        "客户号": 1,
        "行情地址": "",
        "行情端口": 0,
        "交易地址": "",
        "交易端口": 0,
        "行情协议": ["TCP", "UDP"],
        "日志级别": ["FATAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE"],
        "授权码": ""
    }

    exchanges: List[Exchange] = list(EXCHANGE_VT2XTP.keys())

    def __init__(self, event_engine: EventEngine):
        """构造函数"""
        super().__init__(event_engine, "XTP")

        self.md_api = XtpMdApi(self)
        self.td_api = XtpTdApi(self)

    def connect(self, setting: dict) -> None:
        """连接交易接口"""
        userid = setting["账号"]
        password = setting["密码"]
        client_id = int(setting["客户号"])
        quote_ip = setting["行情地址"]
        quote_port = int(setting["行情端口"])
        trader_ip = setting["交易地址"]
        trader_port = int(setting["交易端口"])
        quote_protocol = setting["行情协议"]
        log_level = LOGLEVEL_VT2XTP[setting["日志级别"]]
        software_key = setting["授权码"]

        self.md_api.connect(
            userid, password, client_id, quote_ip,
            quote_port, quote_protocol, log_level
        )
        self.td_api.connect(
            userid, password, client_id, trader_ip,
            trader_port, software_key, log_level
        )
        self.init_query()

    def close(self) -> None:
        """关闭接口"""
        pass

    def subscribe(self, req: SubscribeRequest) -> None:
        """订阅行情"""
        self.md_api.subscribe(req)

    def send_order(self, req: OrderRequest) -> str:
        """委托下单"""
        return self.td_api.send_order(req)

    def cancel_order(self, req: CancelRequest) -> None:
        """委托撤单"""
        self.td_api.cancel_order(req)

    def query_account(self) -> None:
        """查询资金"""
        self.td_api.query_account()

    def query_position(self) -> None:
        """查询持仓"""
        self.td_api.query_position()

    def process_timer_event(self, event) -> None:
        """定时事件处理"""
        self.count += 1
        if self.count < 2:
            return
        self.count = 0

        func = self.query_functions.pop(0)
        func()
        self.query_functions.append(func)

    def init_query(self) -> None:
        """初始化查询任务"""
        self.count = 0
        self.query_functions = [self.query_account, self.query_position]
        self.event_engine.register(EVENT_TIMER, self.process_timer_event)

    def write_error(self, msg: str, error: dict) -> None:
        """输出错误信息"""
        error_id = error["error_id"]
        error_msg = error["error_msg"]
        msg = f"{msg}，代码：{error_id}，信息：{error_msg}"
        self.write_log(msg)


class XtpMdApi(MdApi):

    def __init__(self, gateway: XtpGateway):
        """构造函数"""
        super().__init__()

        self.gateway: XtpGateway = gateway
        self.gateway_name: str = gateway.gateway_name

        self.userid: str = ""
        self.password: str = ""
        self.client_id: int = 0
        self.server_ip: str = ""
        self.server_port: int = 0
        self.protocol: int = 0
        self.session_id: int = 0

        self.connect_status: bool = False
        self.login_status: bool = False

        self.sse_inited: bool = False
        self.szse_inited: bool = False

    def onDisconnected(self, reason: int) -> None:
        """服务器连接断开回报"""
        self.connect_status = False
        self.login_status = False
        self.gateway.write_log(f"行情服务器连接断开, 原因{reason}")

        self.login_server()

    def onError(self, error: dict) -> None:
        """请求报错回报"""
        self.gateway.write_error("行情接口报错", error)

    def onSubMarketData(self, data: dict, error: dict, last: bool) -> None:
        """订阅行情回报"""
        if not error or not error["error_id"]:
            return

        self.gateway.write_error("行情订阅失败", error)

    def onUnSubMarketData(self, data: dict, error: dict, last: bool) -> None:
        """"""
        pass

    def onDepthMarketData(self, data: dict) -> None:
        """行情推送回报"""
        timestamp = str(data["data_time"])
        dt = datetime.strptime(timestamp, "%Y%m%d%H%M%S%f")
        dt = CHINA_TZ.localize(dt)

        tick = TickData(
            symbol=data["ticker"],
            exchange=EXCHANGE_XTP2VT[data["exchange_id"]],
            datetime=dt,
            volume=data["qty"],
            last_price=data["last_price"],
            limit_up=data["upper_limit_price"],
            limit_down=data["lower_limit_price"],
            open_price=data["open_price"],
            high_price=data["high_price"],
            low_price=data["low_price"],
            pre_close=data["pre_close_price"],
            gateway_name=self.gateway_name
        )

        tick.bid_price_1, tick.bid_price_2, tick.bid_price_3, tick.bid_price_4, tick.bid_price_5 = data["bid"][0:5]
        tick.ask_price_1, tick.ask_price_2, tick.ask_price_3, tick.ask_price_4, tick.ask_price_5 = data["ask"][0:5]
        tick.bid_volume_1, tick.bid_volume_2, tick.bid_volume_3, tick.bid_volume_4, tick.bid_volume_5 = data["bid_qty"][0:5]
        tick.ask_volume_1, tick.ask_volume_2, tick.ask_volume_3, tick.ask_volume_4, tick.ask_volume_5 = data["ask_qty"][0:5]

        # 基于合约的最小价格跳动，对浮点数价格进行四舍五入
        contract: ContractData = symbol_contract_map.get(tick.vt_symbol, None)
        if contract:
            pricetick: float = contract.pricetick
            tick.last_price = round_to(data["last_price"], pricetick)
            tick.limit_up = round_to(data["upper_limit_price"], pricetick)
            tick.limit_down = round_to(data["lower_limit_price"], pricetick)
            tick.open_price = round_to(data["open_price"], pricetick)
            tick.high_price = round_to(data["high_price"], pricetick)
            tick.low_price = round_to(data["low_price"], pricetick)
            tick.pre_close = round_to(data["pre_close_price"], pricetick)

            tick.bid_price_1 = round_to(tick.bid_price_1, pricetick)
            tick.bid_price_2 = round_to(tick.bid_price_2, pricetick)
            tick.bid_price_3 = round_to(tick.bid_price_3, pricetick)
            tick.bid_price_4 = round_to(tick.bid_price_4, pricetick)
            tick.bid_price_5 = round_to(tick.bid_price_5, pricetick)
            tick.ask_price_1 = round_to(tick.ask_price_1, pricetick)
            tick.ask_price_2 = round_to(tick.ask_price_2, pricetick)
            tick.ask_price_3 = round_to(tick.ask_price_3, pricetick)
            tick.ask_price_4 = round_to(tick.ask_price_4, pricetick)
            tick.ask_price_5 = round_to(tick.ask_price_5, pricetick)

            tick.name = contract.name

        self.gateway.on_tick(tick)

    def onQueryAllTickers(self, data: dict, error: dict, last: bool) -> None:
        """查询合约回报"""
        contract = ContractData(
            symbol=data["ticker"],
            exchange=EXCHANGE_XTP2VT[data["exchange_id"]],
            name=data["ticker_name"],
            product=PRODUCT_XTP2VT[data["ticker_type"]],
            size=1,
            pricetick=data["price_tick"],
            min_volume=data["buy_qty_unit"],
            gateway_name=self.gateway_name
        )

        if contract.product != Product.OPTION:
            self.gateway.on_contract(contract)

        symbol_contract_map[contract.vt_symbol] = contract

        if last:
            self.gateway.write_log(f"{contract.exchange.value}合约信息查询成功")

            if contract.exchange == Exchange.SSE:
                self.sse_inited = True
            else:
                self.szse_inited = True

            # # 如果上海和深圳都查询完成后，再查询期权的信息
            # if self.sse_inited and self.szse_inited:
            #     self.gateway.td_api.query_option_info()

    def connect(
        self,
        userid: str,
        password: str,
        client_id: int,
        server_ip: str,
        server_port: int,
        quote_protocol: int,
        log_level: int
    ) -> None:
        """连接服务器"""
        self.userid = userid
        self.password = password
        self.client_id = client_id
        self.server_ip = server_ip
        self.server_port = server_port
        self.protocol = PROTOCOL_VT2XTP[quote_protocol]

        # Create API object
        if not self.connect_status:
            path = str(get_folder_path(self.gateway_name.lower()))
            self.createQuoteApi(self.client_id, path, log_level)
            self.login_server()
        else:
            self.gateway.write_log("行情接口已登录，请勿重复操作")

    def login_server(self) -> None:
        """用户登录"""
        n = self.login(
            self.server_ip,
            self.server_port,
            self.userid,
            self.password,
            self.protocol,
            ""
        )

        if not n:
            self.connect_status = True
            self.login_status = True
            msg = "行情服务器登录成功"
            self.query_contract()
            self.init()
        else:
            msg = f"行情服务器登录失败，原因：{n}"
            d = self.getApiLastError()

        self.gateway.write_log(msg)

    def close(self) -> None:
        """关闭连接"""
        if self.connect_status:
            self.exit()

    def subscribe(self, req: SubscribeRequest) -> None:
        """订阅行情"""
        if self.login_status:
            xtp_exchange = EXCHANGE_VT2XTP.get(req.exchange, "")
            self.subscribeMarketData(req.symbol, 1, xtp_exchange)

    def query_contract(self) -> None:
        """查询合约信息"""
        for exchange_id in EXCHANGE_XTP2VT.keys():
            self.queryAllTickers(exchange_id)


class XtpTdApi(TdApi):

    def __init__(self, gateway: XtpGateway):
        """"""
        super().__init__()

        self.gateway: XtpGateway = gateway
        self.gateway_name: str = gateway.gateway_name

        self.userid: str = ""
        self.password: str = ""
        self.client_id: str = ""
        self.server_ip: str = ""
        self.server_port: str = ""
        self.software_key: str = ""

        self.session_id: int = 0
        self.reqid: int = 0
        self.protocol: int = 0

        # 账户是否支持两融或者期权交易
        self.margin_trading = False
        self.option_trading = False

        self.connect_status: bool = False
        self.login_status: bool = False

        self.short_positions: Dict[str, PositionData] = {}
        self.orders: Dict[str, OrderData] = {}

    def onDisconnected(self, session: int, reason: int) -> None:
        """服务器连接断开回报"""
        self.connect_status = False
        self.login_status = False
        self.gateway.write_log(f"交易服务器连接断开, 原因{reason}")

        self.login_server()

    def onError(self, error: dict) -> None:
        """请求报错回报"""
        self.gateway.write_error("交易接口报错", error)

    def onOrderEvent(self, data: dict, error: dict, session: int) -> None:
        """委托状态更新"""
        if error["error_id"]:
            self.gateway.write_error("交易委托失败", error)

        symbol = data["ticker"]
        if len(symbol) == 8:
            direction = DIRECTION_OPTION_XTP2VT[data["side"]]
            offset = OFFSET_XTP2VT[data["position_effect"]]
            order_type = OPTION_ORDERTYPE_XTP2VT.get(data["price_type"], OrderType.MARKET)
        else:
            direction, offset = DIRECTION_STOCK_XTP2VT[data["side"]]

            if symbol.startswith("688"):
                type_map = STAR_ORDERTYPE_XTP2VT
            else:
                type_map = EQUITY_ORDERTYPE_XTP2VT
            order_type = type_map.get(data["price_type"], OrderType.MARKET)

        orderid = str(data["order_xtp_id"])
        if orderid not in self.orders:
            order = OrderData(
                symbol=symbol,
                exchange=MARKET_XTP2VT[data["market"]],
                orderid=orderid,
                type=order_type,
                direction=direction,
                offset=offset,
                price=data["price"],
                volume=data["quantity"],
                traded=data["qty_traded"],
                status=STATUS_XTP2VT[data["order_status"]],
                gateway_name=self.gateway_name
            )
            self.orders[orderid] = order
        else:
            order = self.orders[orderid]
            order.traded = data["qty_traded"]
            order.status = STATUS_XTP2VT[data["order_status"]]

        if not order.datetime:
            timestamp = str(data["insert_time"])
            dt = datetime.strptime(timestamp, "%Y%m%d%H%M%S%f")
            dt = CHINA_TZ.localize(dt)
            order.datetime = dt

        self.gateway.on_order(copy(order))

    def onTradeEvent(self, data: dict, session: int) -> None:
        """成交推送"""
        symbol = data["ticker"]
        if len(symbol) == 8:
            direction = DIRECTION_OPTION_XTP2VT[data["side"]]
            offset = OFFSET_XTP2VT[data["position_effect"]]
        else:
            direction, offset = DIRECTION_STOCK_XTP2VT[data["side"]]

        timestamp = str(data["trade_time"])
        dt = datetime.strptime(timestamp, "%Y%m%d%H%M%S%f")
        dt = CHINA_TZ.localize(dt)

        trade = TradeData(
            symbol=symbol,
            exchange=MARKET_XTP2VT[data["market"]],
            orderid=str(data["order_xtp_id"]),
            tradeid=str(data["exec_id"]),
            direction=direction,
            offset=offset,
            price=data["price"],
            volume=data["quantity"],
            datetime=dt,
            gateway_name=self.gateway_name
        )

        if trade.orderid in self.orders:
            order = self.orders[trade.orderid]
            order.traded += trade.volume

            if order.traded < order.volume:
                order.status = Status.PARTTRADED
            else:
                order.status = Status.ALLTRADED
            self.gateway.on_order(copy(order))
        else:
            self.gateway.write_log(f"成交找不到对应委托{trade.orderid}")

        self.gateway.on_trade(trade)

    def onCancelOrderError(self, data: dict, error: dict, session: int) -> None:
        """撤单报错"""
        if not error or not error["error_id"]:
            return

        self.gateway.write_error("撤单失败", error)

    def onQueryPosition(
        self,
        data: dict,
        error: dict,
        request: int,
        last: bool,
        session: int
    ) -> None:
        """查询持仓回报"""
        if data["market"] == 0:
            return

        position = PositionData(
            symbol=data["ticker"],
            exchange=MARKET_XTP2VT[data["market"]],
            direction=POSITION_DIRECTION_XTP2VT[data["position_direction"]],
            volume=data["total_qty"],
            frozen=data["total_qty"] - data["sellable_qty"],
            price=data["avg_price"],
            pnl=data["unrealized_pnl"],
            yd_volume=data["yesterday_position"],
            gateway_name=self.gateway_name
        )
        self.gateway.on_position(position)

    def onQueryAsset(
        self,
        data: dict,
        error: dict,
        request: int,
        last: bool,
        session: int
    ) -> None:
        """查询资金回报"""
        account = AccountData(
            accountid=self.userid,
            balance=round(data["total_asset"], 2),
            frozen=round(data["withholding_amount"], 2),
            gateway_name=self.gateway_name
        )
        account.available = round(data["buying_power"], 2)

        if data["account_type"] == 1:
            self.margin_trading = True
        elif data["account_type"] == 2:
            account.frozen = round(account.balance - account.available, 2)
            self.option_trading = True

        self.gateway.on_account(account)

    def onQueryOptionAuctionInfo(self, data: dict, error: dict, reqid: int, last: bool, session: int) -> None:
        """查询期权合约细节回报"""
        if not data or not data["ticker"]:
            return

        contract = ContractData(
            symbol=data["ticker"],
            exchange=MARKET_XTP2VT[data["security_id_source"]],
            name=data["symbol"],
            product=Product.OPTION,
            size=data["contract_unit"],
            min_volume=data["qty_unit"],
            pricetick=data["price_tick"],
            gateway_name=self.gateway_name
        )

        contract.option_portfolio = data["underlying_security_id"] + "_O"
        contract.option_underlying = (
            data["underlying_security_id"]
            + "-"
            + str(data["delivery_month"])
        )
        contract.option_type = OPTIONTYPE_XTP2VT.get(data["call_or_put"], None)

        contract.option_strike = data["exercise_price"]
        contract.option_expiry = datetime.strptime(
            str(data["last_trade_date"]), "%Y%m%d"
        )
        contract.option_index = get_option_index(
            contract.option_strike, data["contract_id"]
        )

        self.gateway.on_contract(contract)
        symbol_contract_map[contract.vt_symbol] = contract

        if last:
            self.gateway.write_log("期权信息查询成功")

    def onQueryCreditDebtInfo(
        self,
        data: dict,
        error: dict,
        request: int,
        last: bool,
        session: int
    ) -> None:
        """查询两融持仓回报"""
        if data["debt_type"] == 1:
            symbol = data["ticker"]
            exchange = MARKET_XTP2VT[data["market"]]

            position = self.short_positions.get(symbol, None)
            if not position:
                position = PositionData(
                    symbol=symbol,
                    exchange=exchange,
                    direction=Direction.SHORT,
                    gateway_name=self.gateway_name
                )
                self.short_positions[symbol] = position

            position.volume += data["remain_qty"]

        if last:
            for position in self.short_positions.values():
                self.gateway.on_position(position)

            self.short_positions.clear()

    def connect(
        self,
        userid: str,
        password: str,
        client_id: int,
        server_ip: str,
        server_port: int,
        software_key: str,
        log_level: int
    ) -> None:
        """连接服务器"""

        self.userid = userid
        self.password = password
        self.client_id = client_id
        self.server_ip = server_ip
        self.server_port = server_port
        self.software_key = software_key
        self.protocol = PROTOCOL_VT2XTP["TCP"]

        # Create API object
        if not self.connect_status:
            path = str(get_folder_path(self.gateway_name.lower()))
            self.createTraderApi(self.client_id, path, log_level)

            self.setSoftwareKey(self.software_key)
            self.subscribePublicTopic(0)
            self.login_server()
        else:
            self.gateway.write_log("交易接口已登录，请勿重复操作")

    def login_server(self) -> None:
        """登录"""
        n = self.login(
            self.server_ip,
            self.server_port,
            self.userid,
            self.password,
            self.protocol
        )

        if n:
            self.session_id = n
            self.connect_status = True
            self.login_status = True
            msg = f"交易服务器登录成功, 会话编号：{self.session_id}"
            self.init()
        else:
            error = self.getApiLastError()
            msg = f"交易服务器登录失败，原因：{error['error_msg']}"

        self.gateway.write_log(msg)
        self.query_option_info()

    def close(self) -> None:
        """关闭连接"""
        if self.connect_status:
            self.exit()

    def query_option_info(self) -> None:
        """查询期权信息"""
        self.reqid += 1
        self.queryOptionAuctionInfo({}, self.session_id, self.reqid)

    def send_order(self, req: OrderRequest) -> str:
        """委托下单"""
        if req.exchange not in MARKET_VT2XTP:
            self.gateway.write_log(f"委托失败，不支持的交易所{req.exchange.value}")
            return ""

        if self.margin_trading and req.offset == Offset.NONE:
            self.gateway.write_log("委托失败，两融交易需要选择开平方向")
            return ""

        # 期权委托
        if len(req.symbol) == 8:
            if req.type not in OPTION_ORDERTYPE_VT2XTP:
                self.gateway.write_log(f"委托失败，不支持的期权委托类型{req.type.value}")
                return ""

            xtp_req = {
                "ticker": req.symbol,
                "market": MARKET_VT2XTP[req.exchange],
                "price": req.price,
                "quantity": int(req.volume),
                "side": DIRECTION_OPTION_VT2XTP.get(req.direction, ""),
                "position_effect": OFFSET_VT2XTP[req.offset],
                "price_type": OPTION_ORDERTYPE_VT2XTP[req.type],
                "business_type": 10
            }
        # 股票委托
        else:
            # 科创版
            if req.symbol.startswith("688"):
                type_map = STAR_ORDERTYPE_VT2XTP
            # 其他
            else:
                type_map = EQUITY_ORDERTYPE_VT2XTP

            if req.type not in type_map:
                self.gateway.write_log(f"委托失败，不支持的股票委托类型{req.type.value}")
                return ""

            xtp_req = {
                "ticker": req.symbol,
                "market": MARKET_VT2XTP[req.exchange],
                "price": req.price,
                "quantity": int(req.volume),
                "price_type": type_map[req.type],
            }

            if self.margin_trading:
                xtp_req["side"] = DIRECTION_STOCK_VT2XTP.get((req.direction, req.offset), "")
                xtp_req["business_type"] = 4
            else:
                xtp_req["side"] = DIRECTION_STOCK_VT2XTP.get((req.direction, Offset.NONE), "")
                xtp_req["business_type"] = 0

        orderid = self.insertOrder(xtp_req, self.session_id)
        orderid = str(orderid)

        order = req.create_order_data(orderid, self.gateway_name)
        self.orders[orderid] = order
        self.gateway.on_order(copy(order))

        return order.vt_orderid

    def cancel_order(self, req: CancelRequest) -> None:
        """委托撤单"""
        self.cancelOrder(int(req.orderid), self.session_id)

    def query_account(self) -> None:
        """查询资金"""
        if not self.connect_status:
            return

        self.reqid += 1
        self.queryAsset(self.session_id, self.reqid)

    def query_position(self) -> None:
        """查询持仓"""
        if not self.connect_status:
            return

        self.reqid += 1
        self.queryPosition("", self.session_id, self.reqid)

        if self.margin_trading:
            self.reqid += 1
            self.queryCreditDebtInfo(self.session_id, self.reqid)


def get_option_index(strike_price: float, exchange_instrument_id: str) -> str:
    """获取期权索引"""
    exchange_instrument_id = exchange_instrument_id.replace(" ", "")

    if "M" in exchange_instrument_id:
        n = exchange_instrument_id.index("M")
    elif "A" in exchange_instrument_id:
        n = exchange_instrument_id.index("A")
    elif "B" in exchange_instrument_id:
        n = exchange_instrument_id.index("B")
    else:
        return str(strike_price)

    index = exchange_instrument_id[n:]
    option_index = f"{strike_price:.3f}-{index}"

    return option_index
